import Foundation

final class LocalizedWebsiteGenerator {

    private let content: Content

    let language: ContentLanguage

    private let imageGenerator: ImageGenerator

    private let localizedPostSettings: LocalizedPostSettings

    init(content: Content, language: ContentLanguage) {
        self.language = language
        self.content = content
        self.localizedPostSettings = content.settings.localized(in: language)
        self.imageGenerator = ImageGenerator(
            storage: content.storage,
            settings: content.settings)
        self.outputDirectory = content.storage.outputPath!
    }

    private let outputDirectory: URL

    private var postsPerPage: Int {
        content.settings.posts.postsPerPage
    }

    private var mainContentMaximumWidth: CGFloat {
        CGFloat(content.settings.posts.contentWidth)
    }

    func generateWebsite(callback: (String) -> Void) -> Bool {
        guard imageGenerator.prepareForGeneration() else {
            return false
        }
        guard createMainPostFeedPages() else {
            return false
        }
        #warning("Generate content pages")
        #warning("Generate tag overview page")
        guard generateTagPages() else {
            return false
        }
        guard imageGenerator.runJobs(callback: callback) else {
            return false
        }
        return imageGenerator.save()
    }

    private func createMainPostFeedPages() -> Bool {
        let generator = PostListPageGenerator(
            language: language,
            content: content,
            imageGenerator: imageGenerator,
            showTitle: false,
            pageTitle: localizedPostSettings.title,
            pageDescription: localizedPostSettings.description,
            pageUrlPrefix: localizedPostSettings.feedUrlPrefix)
        return generator.createPages(for: content.posts)
    }

    private func generateTagPages() -> Bool {
        for tag in content.tags {
            let posts = content.posts.filter { $0.tags.contains(tag) }
            guard posts.count > 0 else { continue }

            let localized = tag.localized(in: language)

            let urlPrefix = content.absoluteUrlPrefixForTag(tag, language: language)

            let generator = PostListPageGenerator(
                language: language,
                content: content,
                imageGenerator: imageGenerator,
                showTitle: true,
                pageTitle: localized.name,
                pageDescription: localized.description ?? "",
                pageUrlPrefix: urlPrefix)
            guard generator.createPages(for: posts) else {
                return false
            }
        }
        return true
    }

    private func copy(requiredFiles: Set<FileResource>) -> Bool {
        //print("Copying \(requiredVideoFiles.count) files...")
        for file in requiredFiles {
            guard !file.isExternallyStored else {
                continue
            }

            do {
                try content.storage.copy(file: file.id, to: file.absoluteUrl)
            } catch {
                print("Failed to copy file \(file.id): \(error)")
                return false
            }
        }
        return true
    }

    private func save(_ content: String, to relativePath: String) -> Bool {
        do {
            try self.content.storage.write(content: content, to: relativePath)
            return true
        } catch {
            print("Failed to write page \(relativePath)")
            return false
        }
    }
}
